#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.app.firebase.exception
      @file: exceptions.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.core.exception.exceptions import HSBaseException


class FirebaseAuthenticationError(HSBaseException):
    """Raised when authenticate to Firebase"""


class InvalidFirebaseCredentials(HSBaseException):
    """Raised when invalid credentials are provided to Firebase"""


class FirebaseException(HSBaseException):
    """Raised when Firebase module raises an error"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.app.firebase.src.main
      @file: __main__.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import logging as log
import os
import sys
from textwrap import dedent

import urllib3
from hspylib.core.enums.charset import Charset
from hspylib.core.tools.commons import syserr
from hspylib.core.tools.zoned_datetime import now
from hspylib.modules.cli.application.application import Application
from hspylib.modules.cli.application.version import Version

from firebase.__classpath__ import _Classpath
from firebase.core.firebase import Firebase

# Disable this warning because we are hitting our own database
urllib3.disable_warnings()


class Main(Application):
    """Firebase Agent - Manage your firebase integration"""

    # The welcome message
    DESCRIPTION = _Classpath.get_source_path("welcome.txt").read_text(encoding=str(Charset.UTF_8))

    # location of the .version file
    VERSION_DIR = _Classpath.source_root()

    # The resources folder
    RESOURCE_DIR = str(_Classpath.resource_dir())

    def __init__(self, app_name: str):
        version = Version.load(load_dir=self.VERSION_DIR)
        super().__init__(app_name, version, self.DESCRIPTION.format(version), resource_dir=self.RESOURCE_DIR)
        self.firebase = Firebase()

    def _setup_arguments(self) -> None:
        # @formatter:off
        self._with_options() \
            .option(
                'config-dir', 'd', 'config-dir',
                "the configuration directory. If omitted, the User's home will be used.",
                nargs='?', default=os.getenv('HOME', self._run_dir))
        self._with_chained_args('operation', 'the Firebase operation to process') \
            .argument('setup', 'setup your Firebase account') \
            .argument('upload', 'upload files to your Firebase Realtime Database') \
                .add_option(
                    'dest-dir', 'd', 'dest-dir',
                    'the destination directory. If omitted, the current directory will be used.',
                    nargs='?', default=self._run_dir) \
            .add_option(
                    'glob', 'g', 'glob',
                    'filter path names matching a specified glob pattern.',
                    nargs='?') \
            .add_argument('db_alias', 'alias to identify the firebase object to fetch') \
                .add_argument('files', 'list of files to upload (separated by a space)', nargs='*') \
            .argument('download', 'download files from your Firebase Realtime Database') \
                .add_option(
                    'dest-dir', 'd', 'dest-dir',
                    'the destination directory. If omitted, the current directory will be used.',
                    nargs='?', default=self._run_dir) \
            .add_argument('db_alias', 'alias to identify the firebase object to fetch') \
        # @formatter:on

    def _main(self, *params, **kwargs) -> int:
        """Run the application with the command line arguments"""
        log.info(dedent(f'''
        {self._app_name} v{self._app_version}
        Settings ==============================
                STARTED: {now("%Y-%m-%d %H:%M:%S")}
        '''))
        self._exec_application()
        return 0

    def _exec_application(self) -> None:
        """Execute the specified firebase operation"""
        op = self.get_arg('operation')
        if op == 'setup' or not self.firebase.is_configured():
            self.firebase.setup()
        # Already handled above
        if op == 'setup':
            log.debug('Operation is setup but it was already handled')
        elif op == 'upload':
            self.firebase.upload(
                self.get_arg('db_alias'),
                self.get_arg('files'),
                self.get_arg('glob'))
        elif op == 'download':
            self.firebase.download(
                self.get_arg('db_alias'),
                self.get_arg('dest-dir')
            )
        else:
            syserr(f'### Unhandled operation: {op}')
            self.usage(1)


if __name__ == "__main__":
    # Application entry point
    Main('firebase').INSTANCE.run(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.app.firebase.entity
      @file: file_entry.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import base64
import logging as log
import os

from hspylib.core.enums.charset import Charset
from hspylib.core.preconditions import check_state


class FileEntry:
    """TODO"""

    @staticmethod
    def of(file_path: str, file_data: bytes, expected_size: int) -> 'FileEntry':
        """Create a new file entry with the specified contents and expected size"""
        file_entry = FileEntry(file_path)
        file_entry.data = file_data
        file_entry.decode()
        file_entry.size = len(file_entry.data)
        check_state(
            file_entry.size == expected_size,
            f"Received data and expected data sizes mismatch: {file_entry.size} vs {expected_size}")

        return file_entry

    def __init__(self, file_path: str):
        self.path = file_path
        self.size = 0
        self.data = None
        self.load()

    def __str__(self) -> str:
        return '{"path" : "' + self.path + '", "size" : ' + str(self.size) + ', "data" : "' + self.data + '"}'

    def load(self) -> 'FileEntry':
        """TODO"""
        if os.path.exists(self.path):
            with open(self.path, 'r', encoding='utf-8') as f_in:
                self.data = f_in.read()
                self.size = len(self.data)
                log.warning("File \"%s\" is empty", self.path)

        return self

    def save(self) -> 'FileEntry':
        """Write current file data into this file"""
        with open(self.path, 'w+', encoding='utf-8') as f_in:
            f_in.write(self.data)
            f_in.flush()

        return self

    def encode(self) -> 'FileEntry':
        """B64 Encode this file's data"""
        self.data = str(
            base64.urlsafe_b64encode(self.data.encode(str(Charset.UTF_8))), encoding=str(Charset.UTF_8))
        return self

    def decode(self) -> 'FileEntry':
        """B64 Decode this file's data"""
        self.data = str(
            base64.urlsafe_b64decode(self.data), encoding=str(Charset.UTF_8))
        return self



# _*_ coding: utf-8 _*_
#
# hspylib-firebase v0.9.74
#
# Package: main.firebase
"""Package initialization."""

__all__ = [
  '__classpath__', 
  '__main__', 
  'core', 
  'entity', 
  'exception'
]
__version__ = '0.9.74'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.app.firebase.entity
      @file: firebase_entry.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import json
from typing import List

from hspylib.core.enums.charset import Charset
from hspylib.core.tools.zoned_datetime import now

from firebase.entity.file_entry import FileEntry


class FirebaseEntry:

    def __init__(
        self,
        name: str = None,
        last_update_user: str = None,
        files: List[FileEntry] = None,
        encoding: Charset = Charset.UTF_8):
        self.name = name
        self.last_update_user = last_update_user
        self.files = files if files else []
        self.str_encoding = str(encoding).lower()
        self.last_modified = now("%d/%m/%Y %H:%M:%S")

    def __str__(self):
        return json.dumps(self.__dict__)

    def __repr__(self):
        return str(self)

    def payload(self) -> str:
        """Return a firebase request payload for this entry"""
        return str({
            'encoding': self.str_encoding,
            'file_paths': str(self.files),
            'last_modified': self.last_modified,
            'last_update_user': self.last_update_user
        })



# _*_ coding: utf-8 _*_
#
# hspylib-firebase v0.9.74
#
# Package: main.firebase.core
"""Package initialization."""

__all__ = [
  'agent_config', 
  'file_processor', 
  'firebase', 
  'firebase_auth'
]
__version__ = '0.9.74'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.app.firebase.core
      @file: firebase.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import logging as log
import os
from typing import List

from hspylib.core.preconditions import check_argument

from firebase.core.agent_config import AgentConfig
from firebase.core.file_processor import FileProcessor
from firebase.core.firebase_auth import FirebaseAuth


class Firebase:
    """Represents the firebase agent and it's functionalities"""

    def __init__(self) -> None:
        self.processor = FileProcessor()
        self.agent_config = AgentConfig()

    def __str__(self):
        return str(self.agent_config)

    def setup(self) -> None:
        """Setup a firebase creating or reading an existing firebase configuration file"""
        self.agent_config.prompt()
        log.debug('New firebase setup: %s', self.agent_config)

    def upload(self, db_alias: str, file_paths: List[str], glob_exp: str) -> bool:
        """Upload files to firebase"""
        self._authenticate()
        url = f"{self.agent_config.url(db_alias)}.json"
        check_argument(len(file_paths) > 0, "Unable to upload file_paths (zero size).")
        log.debug('Uploading files  alias=%s  files=[%s]', db_alias, ','.join(file_paths))
        return self.processor.upload_files(url, file_paths, glob_exp) > 0

    def download(self, db_alias: str, dest_dir: str) -> bool:
        """Download files from firebase specified by it's aliases"""
        self._authenticate()
        url = f"{self.agent_config.url(db_alias)}.json"
        log.debug('Downloading files  alias=%s  dest_dir=%s', db_alias, dest_dir)
        return self.processor.download_files(url, dest_dir or os.environ.get('HOME')) > 0

    def is_configured(self) -> bool:
        """Checks whether firebase is properly configured or not"""
        return self.agent_config is not None and self.agent_config.firebase_configs is not None

    def _authenticate(self) -> None:
        """TODO"""
        configs = self.agent_config.firebase_configs
        FirebaseAuth.authenticate(configs.project_id, configs.uid)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
      @file: setup.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import pathlib

import setuptools

HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# The version of the package
VERSION = (HERE / "firebase/.version").read_text().strip()

# The package requirements
REQUIREMENTS = list(filter(None, (HERE / "requirements.txt").read_text().splitlines()))

# This call to setup() does all the work
setuptools.setup(
    name='hspylib-firebase',
    version=VERSION,
    description='HomeSetup - Firebase integration',
    author='Hugo Saporetti Junior',
    author_email='yorevs@hotmail.com',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/yorevs/hspylib',
    project_urls={
        'GitHub': 'https://github.com/yorevs/hspylib',
        'PyPi': 'https://pypi.org/project/hspylib-firebase/'
    },
    license='MIT',
    license_files='LICENSE.md',
    packages=setuptools.find_namespace_packages(),
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix"

    ],
    python_requires='>=3.7',
    install_requires=REQUIREMENTS,
    keywords='firebase,google,integration,application',
    platforms='Darwin,Linux'
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.app.firebase.core
      @file: agent_config.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import logging as log
import os
from typing import Optional

from hspylib.core.config.app_config import AppConfigs
from datasource.firebase.firebase_configuration import FirebaseConfiguration
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.tools.commons import file_is_not_empty, sysout
from requests.structures import CaseInsensitiveDict

from firebase.core.firebase_auth import FirebaseAuth
from firebase.exception.exceptions import FirebaseAuthenticationError


class AgentConfig(metaclass=Singleton):
    """Holds the firebase agent configurations"""

    def __init__(self) -> None:
        self.app_configs = AppConfigs.INSTANCE
        self.firebase_configs = None
        if file_is_not_empty(self.filename):
            self._load()
            log.debug('Found and loaded a Firebase configuration: %s', self)

    def __str__(self) -> str:
        return str(self.firebase_configs)

    def __repr__(self):
        return str(self)

    @property
    def hash_str(self) -> str:
        return 'f6680e84-12c0-459c-8cf9-bd469def750d'

    def setup(self, config_dict: CaseInsensitiveDict) -> None:
        """Setup firebase from a dict configuration
        :param config_dict: Firebase configuration dictionary
        """
        user = FirebaseAuth.authenticate(config_dict)
        if user:
            if user.uid != config_dict['UUID']:
                raise FirebaseAuthenticationError(
                    f"Provided UID: {config_dict['UUID']} is different from retrieved UID: {user.uid}")
            config_dict['UUID'] = user.uid
            self.firebase_configs = FirebaseConfiguration.of(config_dict)
            self._save()
        else:
            raise

    def prompt(self) -> None:
        """Create a new firebase configuration by prompting the user for information."""
        config = CaseInsensitiveDict()
        sysout("### Firebase setup")
        sysout('-' * 31)
        config['UUID'] = self.uuid
        config['PROJECT_ID'] = self.project_id
        config['DATABASE'] = self.database
        config['EMAIL'] = self.email
        self.setup(config)

    @property
    def filename(self) -> str:
        """Return the configuration file."""
        file = self.app_configs['hhs.firebase.config.file']
        return file if file else f"{os.getenv('HOME', os.getcwd())}/.firebase"

    @property
    def project_id(self) -> Optional[str]:
        """Return the firebase project ID."""
        pid = self.app_configs['hhs.firebase.project.id']
        return pid if pid else input('Please type your project ID: ')

    @property
    def uuid(self) -> Optional[str]:
        """Return the firebase user ID."""
        uid = self.app_configs['hhs.firebase.user.uid']
        return uid if uid else input('Please type your user UID: ')

    @property
    def email(self) -> Optional[str]:
        """Return the firebase username."""
        em = self.app_configs['hhs.firebase.email']
        return em if em else input('Please type your Email: ')

    @property
    def database(self) -> Optional[str]:
        """Return the firebase project database name."""
        database = self.app_configs['hhs.firebase.database']
        return database if database else input('Please type your database Name: ')

    def url(self, db_alias: str) -> str:
        """Return the firebase project URL"""
        final_alias = db_alias.replace('.', '/')
        return self.firebase_configs.url(final_alias)

    def _load(self) -> None:
        """Load configuration from file"""
        self.firebase_configs = FirebaseConfiguration.of_file(self.filename)

    def _save(self) -> None:
        """Save current firebase configuration"""
        with open(self.filename, 'w+', encoding='utf-8') as f_config:
            f_config.write(str(self))
            sysout(f"Firebase configuration saved => {self.filename} !")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.app.firebase.core
      @file: firebase_auth.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import os
from abc import ABC
from typing import Optional

import firebase_admin
from firebase_admin import auth, credentials
from firebase_admin.auth import UserNotFoundError, UserRecord
from firebase_admin.exceptions import FirebaseError
from hspylib.core.preconditions import check_not_none
from hspylib.core.tools.commons import sysout

from firebase.exception.exceptions import FirebaseAuthenticationError, FirebaseException, InvalidFirebaseCredentials


class FirebaseAuth(ABC):
    """Firebase authentication utils
    Ref: https://www.youtube.com/watch?v=esqNgnayVE8
    """

    @staticmethod
    def _credentials(project_id: str) -> credentials.Certificate:
        """TODO"""

        certificate_file = os.environ.get("HHS_FIREBASE_CERT_FILE")
        check_not_none(certificate_file, project_id)
        try:
            creds = credentials.Certificate(certificate_file.format(project_id=project_id))
        except (IOError, ValueError) as err:
            raise InvalidFirebaseCredentials('Invalid credentials provided') from err

        return creds

    @staticmethod
    def authenticate(project_id: str, uuid: str) -> Optional[UserRecord]:
        """TODO"""

        firebase_admin.initialize_app(FirebaseAuth._credentials(project_id))
        try:
            user = auth.get_user(uuid)
            if user:
                sysout('Firebase authentication succeeded')
                return user
            else:
                raise FirebaseAuthenticationError('Failed to authenticate to Firebase')
        except UserNotFoundError as err:
            raise FirebaseAuthenticationError(f"Failed to authenticate to Firebase => {err}") from err
        except (ValueError, FirebaseError) as err:
            raise FirebaseException(f"An error occurred authenticating Firebase user => {err}") from err



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.app.firebase.core
      @file: file_processor.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import json
import logging as log
import os
from abc import ABC
from fnmatch import fnmatch
from typing import List, Union

from hspylib.core.enums.http_code import HttpCode
from hspylib.core.preconditions import check_argument, check_not_none
from hspylib.core.tools.commons import syserr, sysout
from hspylib.modules.fetch.fetch import get, put
from requests.exceptions import HTTPError

from firebase.entity.file_entry import FileEntry


class FileProcessor(ABC):
    """Utility class to upload and download B64 encoded files"""

    @staticmethod
    def upload_files(url: str, file_paths: List[str], glob_exp: str) -> int:
        """Upload files to URL"""
        data = []
        for f_path in file_paths:
            if os.path.exists(f_path):
                if os.path.isfile(f_path):
                    sysout(f'Uploading file "{f_path}" to Firebase ...')
                    f_entry = FileProcessor._read_and_encode(f_path)
                    data.append(f_entry)
                else:
                    sysout(f'Uploading files from "{f_path}" to Firebase ...')
                    all_files = next(os.walk(f_path))[2]
                    log.debug('\nGlob: %s \nFiles: %s', glob_exp, all_files)
                    for file in all_files:
                        filename = os.path.join(f_path, file)
                        if os.path.isfile(filename) and fnmatch(file, glob_exp or '*.*'):
                            f_entry = FileProcessor._read_and_encode(filename)
                            data.append(f_entry)
            else:
                syserr(f'Input file "{f_path}" does not exist')
        if data:
            payload = FileProcessor._to_json(data)
            response = put(url, payload)
            check_not_none(response)
            if response.status_code != HttpCode.OK:
                raise HTTPError(
                    f'{response.status_code} - Unable to upload into={url} with json_string={payload}')
            paths = ', \n  |- '.join([f.path for f in data])

            sysout(f"%GREEN%File(s) [\n  |- {paths}\n] successfully uploaded to: {url}%NC%")

            return len(data)

        sysout(f'%ORANGE%No files were uploaded from {file_paths} %NC%')

        return 0

    @staticmethod
    def download_files(url: str, destination_dir: str) -> int:
        """Download files from URL"""
        check_argument(
            destination_dir and os.path.exists(destination_dir),
            "Unable find destination directory: {}", destination_dir)
        sysout(f'Downloading files from Firebase into "{destination_dir}" ...')
        response = get(url)
        check_not_none(response)
        check_not_none(response.body)
        if response.status_code != HttpCode.OK:
            raise HTTPError(
                f'{response.status_code} - Unable to download from={url} with response={response}')
        file_data = FileProcessor._from_json(response.body)
        if file_data and len(file_data) > 0:
            FileProcessor._decode_and_write(destination_dir, file_data)
            return len(file_data)

        sysout(f'%ORANGE%Database alias was not found in: {url} %NC%')

        return 0

    @staticmethod
    def _read_and_encode(file_path: str) -> FileEntry:
        """Read and B64 encode a file"""
        return FileEntry(file_path).encode()

    @staticmethod
    def _decode_and_write(destination_dir: str, data: Union[dict, List[dict]]) -> None:
        """B64 decode and write entries to file"""
        if isinstance(data, list):
            for entry in data:
                FileEntry \
                    .of(f"{destination_dir}/{os.path.basename(entry['path'])}", entry['data'], entry['size']).save()
            paths = ', \n  |- '.join([f['path'] for f in data])
        elif isinstance(data, dict):
            FileEntry \
                .of(f"{destination_dir}/{os.path.basename(data['path'])}", data['data'], data['size']).save()
            paths = data['path']
        else:
            raise TypeError(f'Downloaded data format is not supported: {type(data)}')

        sysout(f"%GREEN%File(s) [\n  |- {paths}\n] successfully downloaded into: {destination_dir}%NC%")

    @staticmethod
    def _to_json(file_data: List[FileEntry]) -> str:
        """Convert the file data into json format"""
        return '[' + ','.join([str(entry) for entry in file_data]) + ']'

    @staticmethod
    def _from_json(file_data: str) -> List[dict]:
        """Convert json format into file data"""
        return json.loads(file_data)
